

/*  GetLayoutItemAttrsA.c  $ 31/05/95 MT $  */


#include "exec/types.h"
#include "graphics/text.h"
#include "intuition/intuition.h"
#include "utility/tagitem.h"
#include "proto/utility.h"
#include "rl_fastlayout.h"
#include "fastlayout.h"
#include "fl_private.h"


void ASM GetLayoutAttrsA(R_A0 APTR, R_A1 APTR);


/****** fastlayout.library/GetLayoutItemAttrsA ******************************
*
*   NAME
*       GetLayoutItemAttrsA -- Inquires about the attributes of an item.
*       GetLayoutItemAttrs -- Varargs version of GetLayoutItemAttrsA().
*
*   SYNOPSIS
*       GetLayoutItemAttrsA(layout, layoutitem, taglist)
*                           A0      A1          A2
*
*       void GetLayoutItemAttrsA(APTR, APTR, struct TagItem *);
*
*       GetLayoutItemAttrs(layout, layoutitem, firsttag, ...)
*
*       void GetLayoutItemAttrs(APTR, APTR, Tag, ...);
*
*   FUNCTION
*       This function allows you to inquire about the attributes of a
*       LayoutItem. Not all of the attributes can be read with this function;
*       the attributes you can inquire about are mainly those you also can
*       change with SetLayoutItemAttrsA().
*       You ask the value of a particular attribute by passing its tag
*       whose tagvalue is a pointer to a variable of the appropriate type.
*       After the function call, you can read that variable to find out the
*       value of the required attribute.
*
*   INPUTS
*       layout - pointer to a reference Layout structure, or NULL if you
*                want the defaults to be used for undefined attributes.
*       layoutitem - pointer to the LayoutItem whose attributes are to
*                    be inquired about.
*       taglist - pointer to an array of TagItem pairs.
*
*   TAGS
*       FL_Item (APTR *) - Pointer to the first sub-item of this item, or
*                          NULL if the item has no sub-items. It is incorrect
*                          to pass this tag if the item is not a zone.
*       FL_Prev (APTR *) - Pointer to the previous item in the list, or NULL
*                          if this item is the first one of its list.
*       FL_Next (APTR *) - Pointer to the next item in the list, or NULL if
*                          this item is the last one of its list.
*       FL_AutoSpace (ULONG *) - The fixed or minimum size of the automatic
*                                spacing, or ~0 (-1) if there is no automatic
*                                spacing.
*       FL_AlignItems (ULONG *) - TRUE if the gadgets of this vertical zone
*                                 have the left or right sides aligned, FALSE
*                                 if they don't.
*       FL_GadgetText (STRPTR **) - Pointer to pointer to the string
*                                   associated to this item, or NULL if
*                                   there's no string.
*       FL_TextAttr (struct TextAttr **) - Pointer to the TextAttr structure
*                                          associated to this item, or NULL
*                                          if this item uses the default
*                                          TextAttr of the Layout.
*       FL_FrameWidth (ULONG *) - Width of the frame of the gadget associated
*                                 to this item. This is the width of ONE of
*                                 the two (left and right) vertical borders
*                                 of the gadget.
*       FL_FrameHeight (ULONG *) - Height of the frame of the gadget
*                                  associated to this item. This is the
*                                  height of ONE of the two (top and bottom)
*                                  horizontal borders of the gadget.
*       FL_BorderType (ULONG *) - Type of border of this item.
*                                 It can be one of the following:
*
*                                 BT_NONE - No border.
*                                 BT_SPACE - Invisible border.
*                                 BT_RAISED - The standard gadtools border.
*                                 BT_RECESSED - As above, but recessed.
*                                 BT_LINE - A furrow (recessed ridge).
*
*                                 Note that there no BT_RIDGE type yet. This
*                                 could change in the future if gadtools will
*                                 support a multi-line string-gadget without
*                                 borders.
*       FL_BorderWidth (ULONG *) - Width of the border of this item. If this
*                                  item has no border, zero will be returned.
*       FL_BorderHeight (ULONG *) - Height of the border of this item. If
*                                   this item has no border, zero will be
*                                   returned.
*       FL_Justify (ULONG *) - Type of justification of this item in the
*                              secundary direction, that is the direction
*                              perpendicular to the one of the layout of the
*                              parent item. This can be one of the following:
*
*                              JT_FLUSH - Spread the item using all the
*                                         available space.
*                              JT_LEFTTOP - Place the item to the left (top)
*                                           side of its parent item.
*                              JT_RIGHTBOTTOM - Place the item to the right
*                                               (bottom) side of its parent
*                                               item.
*                              JT_CENTER - Place the item in the center of
*                                          its parent item.*
*       FL_Size (ULONG *) - Fixed absolute size in pixels of this item in the
*                           direction of the layout of its parent item, or
*                           zero if the item is not of fixed size.
*       FL_RelSize (ULONG *) - Size of this item in the direction of its
*                              parent's layout, in relation to the space
*                              available (the size of the parent item minus
*                              the size of all fixed-size items). This is the
*                              "weight" of this item, so the actual size will
*                              depend on the relative size of the other items
*                              with the FL_RelSize tag. This attribute will
*                              be zero if the item is not of variable size.
*       FL_AutoSize (ULONG *) - TRUE if this item adapts its main size to its
*                               contents, FALSE if not.
*       FL_MinSize (ULONG *) - The absolute minimum size this item can reach
*                              in the direction of its parent's layout.
*                              The actual minimum size could be greater, as
*                              you should take into account the minimum size
*                              of the contents of the item.
*       FL_Labels (STRPTR **) - Pointer to a NULL-terminated array of strings
*                               containing the labels this gadget will
*                               display, or NULL if this item won't become
*                               a list-type gadget.
*       FL_ListInfo (struct ListInfo **) - Pointer to the ListInfo structure
*                                          describing the specific attributes
*                                          of this item, or NULL if this item
*                                          won't become a list-type gadget.
*       FL_WidthGroup (ULONG *) - The ideal "class" of width of this item, or
*                                 zero if this item doesn't belong to any
*                                 width class.
*       FL_HeightGroup (ULONG *) - The ideal "class" of height of this item,
*                                  or zero if this item doesn't belong to any
*                                  height class.
*       FL_HorizLevel (ULONG *) - The maximum length of the level displayed
*                               at the left or right side of the Slider
*                               associated to this item, or zero if this
*                               item won't become a slider or if the level
*                               won't be displayed at its left or right side.
*       FL_VertLevel (ULONG *) - The maximum length of the level displayed
*                                at the top or bottom side of the Slider
*                                associated to this item, or zero if this
*                                item won't become a slider or if the level
*                                won't be displayed at its top or bottom
*                                side.
*       FL_IndicatorID (ULONG *) - The ID of the string gadget to be
*                                  displayed under the ListView associated
*                                  to this item.
*       FL_ItemSpec (struct IBox *) - If you pass a pointer to an IBox
*                                     structure with this tag, the structure
*                                     will be filled with the data
*                                     describing position and size of this
*                                     item.
*
*   EXAMPLE
*
*       \* Finds out the TextAttr and size of this item *\
*
*       struct TextAttr *ta;
*       struct IBox ibox;
*
*       GetLayoutItemAttrs(layout,item,FL_TextAttr,&ta,
*                                      FL_ItemSpec,&ibox,TAG_END);
*
*   NOTES
*       If a tag is repeated, only the last instance will be taken into
*       consideration.
*       Some attributes have a default value in the Layout structure; this
*       default value will be returned when the value in the item is
*       undefined. If you don't pass a valid Layout pointer, the default
*       value returned will be the global one of the library for that
*       attribute.
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       SetLayoutItemAttrsA(), CreataLayoutItemA(), GetLayoutAttrsA()
*
****************************************************************************/

void ASM GetLayoutItemAttrsA(R_A0 struct Layout *layout, R_A1 struct LayoutItem *li, R_A2 struct TagItem *taglist)
{
   struct TagItem *ti;
#ifdef RESLIB
   struct TagItem *tptr = taglist;
#endif
   ULONG *value;

   if (!li) return;

   GetLayoutAttrsA(layout,taglist);

#ifdef RESLIB
   while (ti = NextTagItem(&tptr))
#else
   while (ti = NextTagItem(&taglist))
#endif
   {
      if (!(value = (ULONG *)(ti->ti_Data))) continue;

      switch (ti->ti_Tag)
      {
         case FL_Item:
            *value = (ULONG)(li->First);
            break;

         case FL_AutoSpace:
            *value = ((li->Flags & IF_AUTOSPACE)? (ULONG)(li->AutoSpacing): ~0);
            break;

         case FL_GadgetText:
            *value = (ULONG)(li->Text);
            break;

         case FL_TextAttr:
            if (li->TextAttr) *value = (ULONG)(li->TextAttr);
            break;

         case FL_FrameWidth:

            if (!(li->GadgetHDelta & 0x8000))
            {
               *value = (ULONG)(li->GadgetHDelta / 2);
            }

            break;

         case FL_FrameHeight:

            if (!(li->GadgetVDelta & 0x8000))
            {
               *value = (ULONG)(li->GadgetVDelta / 2);
            }

            break;

         case FL_BorderType:
            *value = (ULONG)(li->BorderType);
            break;

         case FL_BorderWidth:

            if (li->BorderSizes[LT_HORIZ] >= 0)
            {
               *value = (ULONG)(li->BorderSizes[LT_HORIZ]);
            }

            if (!(li->BorderType)) *value = 0;
            break;

         case FL_BorderHeight:

            if (li->BorderSizes[LT_VERT] >= 0)
            {
               *value = (ULONG)(li->BorderSizes[LT_VERT]);
            }

            if (!(li->BorderType)) *value = 0;
            break;

         case FL_AlignItems:
            *value = (ULONG)(!!(li->Flags & IF_ALIGNITEMS));
            break;

         case FL_Justify:
            *value = (ULONG)(li->Flags & IF_JCENTER);
            break;

         case FL_Size:
            *value = (ULONG)((li->SizeType == ST_FIXED)? li->Size: 0);
            break;

         case FL_RelSize:
            *value = (ULONG)((li->SizeType == ST_RELSPACE)? li->Size: 0);
            break;

         case FL_AutoSize:
            *value = (ULONG)(((li->SizeType == ST_RELTEXT) || (li->SizeType == ST_RELCONT))? li->Size: 0);
            break;

         case FL_MinSize:
            *value = (ULONG)(li->MinSize);
            break;

         case FL_Labels:
            *value = (ULONG)(li->Labels);
            break;

         case FL_ListInfo:
            *value = (ULONG)(li->ListInfo);
            break;

         case FL_WidthGroup:
            *value = (ULONG)(li->GroupId[0]);
            break;

         case FL_HeightGroup:
            *value = (ULONG)(li->GroupId[1]);
            break;

         case FL_HorizLevel:
            *value = (ULONG)((li->LevelPlace == LT_HORIZ)? li->MaxLevelLen: 0);
            break;

         case FL_VertLevel:
            *value = (ULONG)((li->LevelPlace == LT_VERT)? li->MaxLevelLen: 0);
            break;

         case FL_ItemSpec:
            ((struct IBox *)value)->Left = li->Left;
            ((struct IBox *)value)->Top = li->Top;
            ((struct IBox *)value)->Width = li->Width;
            ((struct IBox *)value)->Height = li->Height;
            break;

         case FL_Prev:
            *value = (ULONG)(li->Prev);
            break;

         case FL_Next:
            *value = (ULONG)(li->Next);
            break;

         case FL_IndicatorID:
            *value = li->IndicatorId;
            break;
      }
   }
}

#ifndef RESLIB

void GetLayoutItemAttrs(struct Layout *layout, struct LayoutItem *li, Tag tag1, ...)
{
   GetLayoutItemAttrsA(layout,li,(struct TagItem *)&tag1);
}

#endif

