

/*  LabelLists.c  $ 31/05/95 MT $  */


#include "exec/types.h"
#include "exec/lists.h"
#include "exec/memory.h"
#include "proto/exec.h"
#include "rl_fastlayout.h"
#include "fastlayout.h"


static void __saveds __asm PutChProc(register __d0 UBYTE ch, register __a3 UBYTE **putchdata)
{
   *(*putchdata)++ = ch;
}

static void SPrintf(UBYTE *buffer, UBYTE *format, ...)
{
   RawDoFmt(format,((UBYTE *)(&format)) + sizeof(UBYTE *),PutChProc,&buffer);
}

/****** fastlayout.library/ModifyLabelList **********************************
*
*   NAME
*       ModifyLabelList -- Changes the contents of an Exec list of strings.
*
*   SYNOPSIS
*       success = ModifyLabelList(list, array, type, number)
*       D0                        A0    A1     D0    D1
*
*       BOOL ModifyLabelList(struct List *, void *, ULONG, ULONG);
*
*   FUNCTION
*       With this function you can modify the contents of a label list
*       created with CreateLabelList().
*       You can pass to the function an array of strings or integers and
*       the list will be modified so that it contains the new array's
*       elements. All old nodes created with previous calls to either
*       CreateLabelList() or ModifyLabelList() will be deallocated, and new
*       ones will be created to contain the new information. All nodes
*       added by hand to the list will be preserved, but after the call
*       they will all reside in the list's head. The advantage of using this
*       function rather than the couple FreeLabelList()/CreateLabelList() is
*       that the main List structure does not have to be deallocated, so
*       some time is saved (and system memory gets less fragmented).
*
*   INPUTS
*       list - pointer to List structure as returned by CreateLabelList().
*       array - pointer to the array of strings or integers. The actual types
*               should be "STRPTR array[]" or "LONG array[]".
*       type - the type of the array passed. The types are LL_STRING and
*              LL_INTEGER. Be careful to specify the correct type!
*       number - the number of elements in the array. In the case of a
*                NULL-terminated string array (and only in this case!) you
*                can pass a number of -1 (or ~0) to indicate that all of
*                the array's elements are to be used.
*
*   RESULT
*       success - TRUE if all went OK, FALSE if some allocation failed. In
*                 this case you can re-try to assign an array to the list,
*                 or free it with FreeLabelList().
*
*   EXAMPLE
*
*       ...
*       STRPTR newstrings[] = { "One", "Two", "Three", "Four", NULL };
*       BOOL ok;
*
*       ok = ModifyLabelList(mylist,newstrings,LL_STRING,-1);
*       if (!ok) prints("Couldn't modify the list!\n");
*       ...
*
*   CAUTION
*       All nodes created by CreateLabelList() and ModifyLabelList() will
*       have a ln_Type of 121, so don't put this type into your custom nodes
*       or FreeLabelList() will attempt to free them, causing a system crash!
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       CreateLabelList(), FreeLabelList(), <exec/lists.h>, <exec/nodes.h>
*
****************************************************************************/

ULONG ASM ModifyLabelList(R_A0 struct List *list, R_A1 void *array, R_D0 ULONG type, R_D1 ULONG num)
{
   struct Node *node, *pred;
   ULONG stringsize, i = 0;
   UBYTE **stringarray = (UBYTE **)array;
   LONG *integerarray = (LONG *)array;
   ULONG ret = TRUE;

   if (!list) return (FALSE);

   node = list->lh_Head;

   while (node->ln_Succ)
   {
      pred = node;
      node = node->ln_Succ;

      if (pred->ln_Type == 121)
      {
         Remove(pred);
         if (pred->ln_Name) FreeVec(pred->ln_Name);
         FreeVec(pred);
      }
   }

   if (!array) return (TRUE);

   while (i < num)
   {
      if (type == LL_STRING)
      {
         if (stringarray[i] == NULL) break;
         stringsize = strlen(stringarray[i]) + 1;
      }
      else
      {
         stringsize = 16;
      }

      if (!(node = AllocVec(sizeof(struct Node),MEMF_CLEAR)))
      {
         ret = FALSE;
         break;
      }

      node->ln_Type = 121;
      AddTail(list,node);

      if (!(node->ln_Name = (UBYTE *)AllocVec(stringsize,MEMF_CLEAR)))
      {
         ret = FALSE;
         break;
      }

      if (type == LL_STRING)
      {
         strcpy(node->ln_Name,stringarray[i]);
      }
      else if (type == LL_INTEGER)
      {
         SPrintf(node->ln_Name,"%ld",integerarray[i]);
      }

      i += 1;
   }

   if (ret == FALSE)
   {
      node = list->lh_Head;

      while (node->ln_Succ)
      {
         pred = node;
         node = node->ln_Succ;

         if (pred->ln_Type == 121)
         {
            Remove(pred);
            if (pred->ln_Name) FreeVec(pred->ln_Name);
            FreeVec(pred);
         }
      }
   }

   return (ret);
}

/****** fastlayout.library/FreeLabelList ************************************
*
*   NAME
*       FreeLabelList -- Deallocates the memory used by a list of strings.
*
*   SYNOPSIS
*       FreeLabelList(list)
*                     A0
*
*       void FreeLabelList(struct List *);
*
*   FUNCTION
*       This function deallocates a list of labels previously created with
*       CreateLabelList(). You shouldn't call this function with your custom
*       lists! You can, however, add your custom nodes to the list; they
*       will be preserved by FreeLabelList(), even if they won't represent
*       a valid list anymore after the call.
*
*   INPUTS
*       list - pointer to the Exec list to be freed. You can safely pass a
*              NULL here.
*
*   CAUTION
*       All nodes created by CreateLabelList() and ModifyLabelList() will
*       have a ln_Type of 121, so don't put this type into your custom nodes
*       or FreeLabelList() will attempt to free them, causing a system crash!
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       CreateLabelList(), ModifyLabelList(), <exec/lists.h>, <exec/nodes.h>
*
****************************************************************************/

void ASM FreeLabelList(R_A0 struct List *list)
{
   struct Node *node, *pred;

   if (list)
   {
      node = list->lh_Head;

      while (node->ln_Succ)
      {
         pred = node;
         node = node->ln_Succ;

         if (pred->ln_Type == 121)
         {
            Remove(pred);
            if (pred->ln_Name) FreeVec(pred->ln_Name);
            FreeVec(pred);
         }
      }

      FreeVec(list);
   }
}

/****** fastlayout.library/CreateLabelList **********************************
*
*   NAME
*       CreateLabelList -- Allocates and initializes an Exec list of labels.
*
*   SYNOPSIS
*       list = CreateLabelList(array, type, number)
*       D0                     A0     D0    D1
*
*       struct List *CreateLabelList(void *, ULONG, ULONG);
*
*   FUNCTION
*       This function allows you to easily create a list of Exec nodes,
*       each with a pointer to a string of your array in its ln_Name field.
*       This is a convenient way to create the list of labels required by
*       the gadtools.library LISTVIEW_KIND gadgets, but it can come useful
*       also in many other circumstances.
*       You can pass two types of array to this function, an array of
*       string pointers (the common case) or an array of integers (useful,
*       for instance, if you have a ListView which will only display numbers,
*       such as the one in the ASL font requester).
*       You also can create an empty list for later use.
*       You are allowed to add your own custom nodes to the list, but be
*       sure *NOT* to assign them a ln_Type of 121, as this is the value
*       used to recognize nodes created by both CreateLabelList() and
*       ModifyLabelList().
*       All your custom nodes will be preserved when modifying or freeing
*       the list, so you're responsible for deallocating them as needed.
*
*   INPUTS
*       array - pointer to the array of strings or integers. The actual types
*               should be "STRPTR array[]" or "LONG array[]". You also can
*               pass a NULL here, in which case an empty list will be created
*               that you can modify later in your program.
*       type - the type of the array passed. The types are LL_STRING and
*              LL_INTEGER. Be careful to specify the correct type!
*       number - the number of elements in the array. In the case of a
*                NULL-terminated string array (and only in this case!) you
*                can pass a number of -1 (or ~0) to indicate that all of
*                the array's elements are to be used.
*
*   RESULT
*       list - pointer to the new list of labels, or NULL if the allocation
*              failed.
*
*   EXAMPLE
*
*       ...
*       STRPTR strings[] = { "Computer", "Monitor", "Printer", NULL };
*       struct List *list;
*
*       list = CreateLabelList(strings,LL_STRING,-1);
*       if (!list) printf("No memory for the list!\n");
*       ...
*
*   NOTES
*       The strings or the numbers of your array are copied into the list
*       so if you change the array you must modify or re-create the list.
*       Also, you are allowed only to read the fields of the list's nodes
*       (except for custom nodes you added by hand): to modify or deallocate
*       the list you should use the appropriate ModifyLabelList() and
*       FreeLabelList() functions.
*
*   CAUTION
*       All nodes created by CreateLabelList() and ModifyLabelList() will
*       have a ln_Type of 121, so don't put this type into your custom nodes
*       or FreeLabelList() will attempt to free them, causing a system crash!
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       ModifyLabelList(), FreeLabelList(), <exec/lists.h>, <exec/nodes.h>
*
****************************************************************************/

struct List * ASM CreateLabelList(R_A0 void *array, R_D0 ULONG type, R_D1 ULONG num)
{
   struct List *list;

   if (list = AllocVec(sizeof(struct List),MEMF_CLEAR))
   {
      list->lh_Head = (struct Node *)&(list->lh_Tail);
      list->lh_TailPred = (struct Node *)&(list->lh_Head);

      if (!ModifyLabelList(list,array,type,num))
      {
         FreeLabelList(list);
         list = NULL;
      }
   }

   return (list);
}


