

/*  ScanLayoutItemTree.c  $ 17/05/95 MT $  */


#include "exec/types.h"
#include "rl_fastlayout.h"
#include "fastlayout.h"
#include "fl_private.h"


/****** fastlayout.library/ScanLayoutItemTree *******************************
*
*   NAME
*       ScanLayoutItemTree -- Performs a pre- or post-order scan of a tree
*                             of LayoutItem structures, calling a custom
*                             function for each of them.
*
*   SYNOPSIS
*       result = ScanLayoutItemTree(root, scantype, function, userdata)
*                                   A0    D0        A1        A2
*
*       ULONG ScanLayoutItemTree(APTR, ULONG, ULONG (*)(APTR, APTR), APTR);
*
*   FUNCTION
*       This function allows you to call a custom function for each of the
*       LayoutItem structures in a tree. For each LayoutItem encountered,
*       your function is called with a pointer to the LayoutItem and a
*       pointer to user data you passed to the ScanLayoutItemTree() function.
*       Your function must then return a value: if it returns zero the scan
*       continues, else the scan stops and your value is returned immediately
*       by ScanLayoutItemTree().
*
*   IMPLEMENTATION
*       This function is, by its own nature, highly recursive. This means
*       that, depending on the amount of stack your custom function uses,
*       there could be a stack overflow in the case of very deep trees.
*       In this case you should increase the stack allocated by your program.
*
*   INPUTS
*       root - pointer to the root of the LayoutItem tree.
*       scantype - type of scan to be performed. It can be SCAN_NODEFIRST,
*                  which performs a sort of pre-order scan (the function is
*                  called first on the node and then on its children) or
*                  SCAN_NODELAST, which performs exactly the opposite.
*       function - pointer to the custom function to be called for each
*                  LayoutItem of the tree. Your function will receive the
*                  following arguments:
*
*                  A0 - pointer to the LayoutItem structure
*                  A1 - pointer to your personal user data (see below)
*
*                  and must return a value in D0.
*       userdata - pointer to user-private data, for use by the custom
*                  function. You can specify whatever you like here, even
*                  a non-pointer value, or NULL. This value is guaranteed
*                  to remain untouched by the library, and is passed to
*                  your custom function every time it is called.
*
*   RESULT
*       result - if during the scan your custom function returns a non-zero
*                value, ScanLayoutItemTree() returns it immediately. If not,
*                the result is zero.
*
*   EXAMPLE
*
*       ULONG PrintItemWidth(APTR layoutitem, APTR filehandle)
*       {
*          struct IBox ibox;
*
*          GetLayoutItemAttrs(layoutitem,FL_ItemSpec,&ibox,TAG_END);
*          FPrintf((BPTR)filehandle,"Width: %d\n",ibox.Width);
*
*          return (0L);
*       }
*
*       ...
*       ScanLayoutItemTree(root,SCAN_NODEFIRST,PrintItemWidth,debugfile);
*       ...
*
*   BUGS
*       None known, but see "implementation" above.
*
*   SEE ALSO
*
****************************************************************************/

ULONG ASM ScanLayoutItemTree(R_A0 struct LayoutItem *item, R_D0 ULONG scantype, R_A1 ULONG ASM (*function)(R_A0 struct LayoutItem *, R_A1 APTR), R_A2 APTR userdata)
{
   struct LayoutItem *li;
   ULONG ret;

   if (item)
   {
      for (li = item; li; li = li->Next)
      {
         if (scantype == SCAN_NODEFIRST) if (ret = function(li,userdata)) return (ret);
         if (ret = ScanLayoutItemTree(li->First,scantype,function,userdata)) return (ret);
         if (scantype == SCAN_NODELAST) if (ret = function(li,userdata)) return (ret);
      }
   }

   return (0L);
}

